import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "features"
test_name = "assertion"
data_type = "lens_light_dev_vaucouleurs"
data_resolution = "lsst"


def make_pipeline(name, phase_folders, non_linear_class=af.MultiNest):

    sersic = af.PriorModel(al.lp.EllipticalSersic)

    # This will lead to pretty weird results

    sersic.add_assertion(sersic.axis_ratio > sersic.intensity)

    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(lens=al.GalaxyModel(redshift=0.5, sersic=sersic)),
        non_linear_class=non_linear_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    # TODO : And even with them not causing errors above, the promise doesnt work.

    phase2 = al.PhaseImaging(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(lens=phase1.result.model.galaxies.lens),
        non_linear_class=non_linear_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.8

    return al.PipelineDataset(name, phase1, phase2)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



from test_autolens.integration.tests.imaging.lens_only import lens_light__hyper
from test_autolens.integration.tests.imaging.runner import run_a_mock


class TestCase:
    def _test__lens_light__hyper(self):
        run_a_mock(lens_light__hyper)



# from os import path
# from shutil import rmtree
#
# import pytest
#
# import autofit as af
# from autofit.optimize.non_linear.mock_nlo import MockOutput
#
# directory = path.dirname(path.realpath(__file__))
# aggregator_directory = "{}/files/aggregator".format(directory)
#
#
# @pytest.fixture(name="aggregator")
# def make_aggregator():
#     yield af.Aggregator(aggregator_directory)
#     rmtree(f"{aggregator_directory}/one")
#     rmtree(f"{aggregator_directory}/two")
#     rmtree(f"{aggregator_directory}/three")
#
#
# def filter_phases(aggregator, folder):
#     return list(
#         filter(
#             lambda ag: "/{}/metadata".format(folder) in ag.file_path, aggregator.phases
#         )
#     )[0]
#
#
# @pytest.fixture(name="one")
# def make_one(aggregator):
#     return filter_phases(aggregator, "one")
#
#
# @pytest.fixture(name="two")
# def make_two(aggregator):
#     return filter_phases(aggregator, "two")
#
#
# @pytest.fixture(name="three")
# def make_three(aggregator):
#     return filter_phases(aggregator, "three")
#
#
# class TestCase:
#     def test__total_phases(self, aggregator):
#         assert len(aggregator.phases) == 3
#
#     def test__optimizer_output(self, one, two, three):
#         assert isinstance(one.output, MockOutput)
#         assert isinstance(two.output, MockOutput)
#         assert isinstance(three.output, MockOutput)
#
#     def test__file_paths(self, one, two, three):
#         assert three.file_path == "{}/three/three/metadata".format(aggregator_directory)
#         assert one.file_path == "{}/one/one/metadata".format(aggregator_directory)
#         assert two.file_path == "{}/two/two/metadata".format(aggregator_directory)
#
#     # def test_attributes(self, one, two, three):
#     #     assert one.pipeline == "pipeline_1"
#     #     assert one.phase == "phase_1"
#     #     print(one.dataset)
#     #     assert one.dataset == "lens_1"
#     #
#     #     assert two.pipeline == "pipeline_2"
#     #     assert two.phase == "phase_1"
#     #     assert two.dataset == "lens_1"
#     #
#     #     assert three.pipeline == "pipeline_1"
#     #     assert three.phase == "phase_2"
#     #     assert three.dataset == "lens_2"
#     #
#     # def test_filter_phases(self, aggregator, one, two, three):
#     #     result = aggregator.phases_with()
#     #     assert len(result) == 3
#     #     assert one in result
#     #     assert two in result
#     #     assert three in result
#     #
#     #     result = aggregator.phases_with(pipeline="pipeline_1")
#     #     assert len(result) == 2
#     #     assert one in result
#     #     assert three in result
#     #
#     #     result = aggregator.phases_with(dataset="lens_2")
#     #     assert [three] == result
#     #
#     #     result = aggregator.phases_with(pipeline="pipeline_2", phase="phase_1")
#     #     assert [two] == result
#
#     def test_phase_model_results(self, one, two, three):
#         assert one.model_results == "results_one"
#         assert two.model_results == "results_two"
#         assert three.model_results == "results_three"
#
#     # def test_header(self, one, two, three):
#     #     assert one.header == "pipeline_1/phase_1/lens_1"
#     #     assert two.header == "pipeline_2/phase_1/lens_1"
#     #     assert three.header == "pipeline_1/phase_2/lens_2"
#     #
#     # def test_aggregator_model_results(self, aggregator):
#     #     assert sorted(aggregator.model_results) == sorted(
#     #         "pipeline_2/phase_1/lens_1\n\n"
#     #         "results_two\n\n"
#     #         "pipeline_1/phase_2/lens_2\n\n"
#     #         "results_three\n\n"
#     #         "pipeline_1/phase_1/lens_1\n\n"
#     #         "results_one"
#     #     )
#     #
#     #     assert sorted(aggregator.filter(phase="phase_1").model_results) == sorted(
#     #         "pipeline_2/phase_1/lens_1\n\n"
#     #         "results_two\n\n"
#     #         "pipeline_1/phase_1/lens_1\n\n"
#     #         "results_one"
#     #     )
#
#     def test_nlo(self, one, two, three):
#         assert one.optimizer is not None
#         assert two.optimizer is not None
#         assert three.optimizer is not None
#
#         assert one.model.priors == two.model.priors
#
#     def test_filter_optimizers(self, aggregator, one, two, three):
#         result = aggregator.optimizer
#
#         assert len(result) == 3
#         assert one.optimizer in result
#         assert two.optimizer in result
#         assert three.optimizer in result
#
#         result = aggregator.filter(pipeline="pipeline_1").optimizer
#         assert len(result) == 2
#         assert one.optimizer in result
#         assert three.optimizer in result



from autoarray.mask.mask import Mask
from autoarray.structures.arrays import Array
from autoarray.structures.grids import (
    Grid,
    GridIrregular,
    GridRectangular,
    GridVoronoi,
    Coordinates,
)
from autoarray.structures.kernel import Kernel
from autoarray.structures.visibilities import Visibilities
from autoarray.structures.arrays import MaskedArray
from autoarray.structures.grids import MaskedGrid
from autoarray.dataset.imaging import Imaging
from autoarray.dataset.interferometer import Interferometer
from autoarray.dataset import preprocess
from autoarray.operators.convolver import Convolver
from autoarray.operators.transformer import TransformerDFT
from autoarray.operators.transformer import TransformerFFT
from autoarray.operators.transformer import TransformerNUFFT
from autoarray.operators.inversion.mappers import mapper
from autoarray.operators.inversion.inversions import inversion
from autoarray.operators.inversion import pixelizations as pix, regularization as reg
from autoarray import conf

from autoastro import dimensions as dim
from autoastro import util
from autoastro.profiles import (
    light_profiles as lp,
    mass_profiles as mp,
    light_and_mass_profiles as lmp,
)
from autoastro.galaxy.galaxy import Galaxy, HyperGalaxy, Redshift
from autoastro.galaxy.galaxy_data import GalaxyData
from autoastro.galaxy.fit_galaxy import FitGalaxy
from autoastro.galaxy.galaxy_model import GalaxyModel
from autoastro.hyper import hyper_data

from autolens.dataset.imaging import MaskedImaging, SimulatorImaging
from autolens.dataset.interferometer import (
    MaskedInterferometer,
    SimulatorInterferometer,
)
from autolens.lens.plane import Plane
from autolens.lens.ray_tracing import Tracer
from autolens import util
from autolens.fit.fit import FitImaging, FitInterferometer
from autolens.fit.fit import FitPositions
from autolens.pipeline import tagging
from autolens.pipeline.phase.abstract import phase
from autolens.pipeline.phase.abstract.phase import AbstractPhase
from autolens.pipeline.phase.extensions import CombinedHyperPhase
from autolens.pipeline.phase.extensions import HyperGalaxyPhase
from autolens.pipeline.phase.extensions.hyper_galaxy_phase import HyperGalaxyPhase
from autolens.pipeline.phase.extensions.hyper_phase import HyperPhase
from autolens.pipeline.phase.extensions.inversion_phase import (
    InversionBackgroundBothPhase,
    InversionBackgroundNoisePhase,
    InversionBackgroundSkyPhase,
    InversionPhase,
    ModelFixingHyperPhase,
)
from autolens.pipeline.phase.abstract.phase import AbstractPhase
from autolens.pipeline.phase.dataset.phase import PhaseDataset
from autolens.pipeline.phase.imaging.phase import PhaseImaging
from autolens.pipeline.phase.interferometer.phase import PhaseInterferometer
from autolens.pipeline.phase.phase_galaxy import PhaseGalaxy
from autolens.pipeline.pipeline import PipelineDataset, PipelinePositions
from autolens.pipeline import setup
from autolens import plot

__version__ = '0.42.0'



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "grid_search"
test_name = "multinest_grid_subhalo__hyper__parallel"
data_type = "lens_sie__source_smooth"
data_resolution = "lsst"


def make_pipeline(name, phase_folders, non_linear_class=af.MultiNest):

    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, mass=al.mp.EllipticalIsothermal),
            source=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
        ),
        non_linear_class=non_linear_class,
    )

    phase2 = al.PhaseImaging(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=phase1.result.instance.galaxies.lens,
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pix.VoronoiBrightnessImage,
                regularization=al.reg.AdaptiveBrightness,
            ),
        ),
        non_linear_class=non_linear_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.8

    phase2 = phase2.extend_with_multiple_hyper_phases(
        hyper_galaxy=False, inversion=False
    )

    class GridPhase(af.as_grid_search(al.PhaseImaging, parallel=True)):
        @property
        def grid_priors(self):
            return [
                self.model.galaxies.subhalo.mass.centre_0,
                self.model.galaxies.subhalo.mass.centre_1,
            ]

    subhalo = al.GalaxyModel(
        redshift=0.5, mass=al.mp.SphericalTruncatedNFWMassToConcentration
    )

    subhalo.mass.mass_at_200 = af.LogUniformPrior(lower_limit=1.0e6, upper_limit=1.0e11)

    subhalo.mass.centre_0 = af.UniformPrior(lower_limit=-2.0, upper_limit=2.0)

    subhalo.mass.centre_1 = af.UniformPrior(lower_limit=-2.0, upper_limit=2.0)

    phase3 = GridPhase(
        phase_name="phase_3",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase2.result.model.galaxies.lens.mass
            ),
            subhalo=subhalo,
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=phase2.result.instance.galaxies.source.pixelization,
                regularization=phase2.result.instance.galaxies.source.regularization,
            ),
        ),
        non_linear_class=non_linear_class,
        number_of_steps=2,
    )

    phase3.optimizer.const_efficiency_mode = True

    return al.PipelineDataset(name, phase1, phase2, phase3)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "grid_search"
test_name = "multinest_grid__subhalo"
data_type = "lens_sie__source_smooth"
data_resolution = "lsst"


def make_pipeline(name, phase_folders, non_linear_class=af.MultiNest):

    lens = al.GalaxyModel(redshift=0.5, mass=al.mp.EllipticalIsothermal)

    lens.mass.centre_0 = af.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
    lens.mass.centre_1 = af.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
    lens.mass.axis_ratio = af.UniformPrior(lower_limit=0.65, upper_limit=0.75)
    lens.mass.phi = af.UniformPrior(lower_limit=40.0, upper_limit=50.0)
    lens.mass.einstein_radius = af.UniformPrior(lower_limit=1.55, upper_limit=1.65)

    source = al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic)

    source.light.centre_0 = af.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
    source.light.centre_1 = af.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
    source.light.axis_ratio = af.UniformPrior(lower_limit=0.75, upper_limit=0.85)
    source.light.phi = af.UniformPrior(lower_limit=50.0, upper_limit=70.0)
    source.light.intensity = af.UniformPrior(lower_limit=0.35, upper_limit=0.45)
    source.light.effective_radius = af.UniformPrior(lower_limit=0.45, upper_limit=0.55)
    source.light.sersic_index = af.UniformPrior(lower_limit=0.9, upper_limit=1.1)

    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(lens=lens, source=source),
        non_linear_class=non_linear_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    class GridPhase(af.as_grid_search(phase_class=al.PhaseImaging, parallel=False)):
        @property
        def grid_priors(self):
            return [
                self.model.galaxies.subhalo.mass.centre_0,
                self.model.galaxies.subhalo.mass.centre_1,
            ]

    subhalo = al.GalaxyModel(
        redshift=0.5, mass=al.mp.SphericalTruncatedNFWMassToConcentration
    )

    subhalo.mass.mass_at_200 = af.LogUniformPrior(lower_limit=1.0e6, upper_limit=1.0e11)

    subhalo.mass.centre_0 = af.UniformPrior(lower_limit=-2.0, upper_limit=2.0)

    subhalo.mass.centre_1 = af.UniformPrior(lower_limit=-2.0, upper_limit=2.0)

    phase2 = GridPhase(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=af.last.instance.galaxies.lens,
            subhalo=subhalo,
            source=af.last.instance.galaxies.source,
        ),
        non_linear_class=non_linear_class,
        number_of_steps=2,
    )

    phase2.optimizer.const_efficiency_mode = True

    phase3 = al.PhaseImaging(
        phase_name="phase_3__subhalo_refine",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=af.last[-1].model.galaxies.lens,
            subhalo=phase2.result.model.galaxies.subhalo,
            source=af.last[-1].instance.galaxies.source,
        ),
        non_linear_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True

    return al.PipelineDataset(name, phase1, phase2, phase3)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



from autoarray.structures import grids
from autolens.fit import fit
from autolens import exc


class AbstractLensMasked:
    def __init__(self, positions, positions_threshold, preload_sparse_grids_of_planes):

        if positions is not None:
            self.positions = grids.Coordinates(coordinates=positions)
        else:
            self.positions = None

        self.positions_threshold = positions_threshold

        self.preload_sparse_grids_of_planes = preload_sparse_grids_of_planes

    def check_positions_trace_within_threshold_via_tracer(self, tracer):

        if self.positions is not None and self.positions_threshold is not None:

            positions_fit = fit.FitPositions(
                positions=self.positions,
                tracer=tracer,
                noise_map=self.imaging.pixel_scales,
            )

            if not positions_fit.maximum_separation_within_threshold(
                self.positions_threshold
            ):
                raise exc.RayTracingException

    def check_inversion_pixels_are_below_limit_via_tracer(self, tracer):

        if self.inversion_pixel_limit is not None:
            pixelizations = list(filter(None, tracer.pixelizations_of_planes))
            if pixelizations:
                for pixelization in pixelizations:
                    if pixelization.pixels > self.inversion_pixel_limit:
                        raise exc.PixelizationException



import os
from os import path

import numpy as np
import pytest
from astropy import cosmology as cosmo

import autofit as af
import autolens as al
from autolens.fit.fit import FitImaging
from test_autolens.mock import mock_pipeline

pytestmark = pytest.mark.filterwarnings(
    "ignore:Using a non-tuple sequence for multidimensional indexing is deprecated; use `arr[tuple(seq)]` instead of "
    "`arr[seq]`. In the future this will be interpreted as an arrays index, `arr[np.arrays(seq)]`, which will result "
    "either in an error or a different result."
)

directory = path.dirname(path.realpath(__file__))


def clean_images():
    try:
        os.remove("{}/source_lens_phase/source_image_0.fits".format(directory))
        os.remove("{}/source_lens_phase/lens_image_0.fits".format(directory))
        os.remove("{}/source_lens_phase/model_image_0.fits".format(directory))
    except FileNotFoundError:
        pass
    af.conf.instance.dataset_path = directory


class TestAttributes:
    def test__masked_imaging_signal_to_noise_limit(self, imaging_7x7, mask_7x7_1_pix):
        imaging_snr_limit = imaging_7x7.signal_to_noise_limited_from_signal_to_noise_limit(
            signal_to_noise_limit=1.0
        )

        phase_imaging_7x7 = al.PhaseImaging(
            phase_name="phase_imaging_7x7", signal_to_noise_limit=1.0
        )

        analysis = phase_imaging_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_7x7_1_pix
        )
        assert (
            analysis.masked_dataset.image.in_2d
            == imaging_snr_limit.image.in_2d * np.invert(mask_7x7_1_pix)
        ).all()
        assert (
            analysis.masked_dataset.noise_map.in_2d
            == imaging_snr_limit.noise_map.in_2d * np.invert(mask_7x7_1_pix)
        ).all()

        imaging_snr_limit = imaging_7x7.signal_to_noise_limited_from_signal_to_noise_limit(
            signal_to_noise_limit=0.1
        )

        phase_imaging_7x7 = al.PhaseImaging(
            phase_name="phase_imaging_7x7", signal_to_noise_limit=0.1
        )

        analysis = phase_imaging_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_7x7_1_pix
        )
        assert (
            analysis.masked_dataset.image.in_2d
            == imaging_snr_limit.image.in_2d * np.invert(mask_7x7_1_pix)
        ).all()
        assert (
            analysis.masked_dataset.noise_map.in_2d
            == imaging_snr_limit.noise_map.in_2d * np.invert(mask_7x7_1_pix)
        ).all()

    def test__masked_imaging_is_binned_up(self, imaging_7x7, mask_7x7_1_pix):
        binned_up_imaging = imaging_7x7.binned_from_bin_up_factor(bin_up_factor=2)

        binned_up_mask = mask_7x7_1_pix.mapping.binned_mask_from_bin_up_factor(
            bin_up_factor=2
        )

        phase_imaging_7x7 = al.PhaseImaging(
            phase_name="phase_imaging_7x7", bin_up_factor=2
        )

        analysis = phase_imaging_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_7x7_1_pix
        )
        assert (
            analysis.masked_dataset.image.in_2d
            == binned_up_imaging.image.in_2d * np.invert(binned_up_mask)
        ).all()

        assert (
            analysis.masked_dataset.psf == (1.0 / 9.0) * binned_up_imaging.psf
        ).all()
        assert (
            analysis.masked_dataset.noise_map.in_2d
            == binned_up_imaging.noise_map.in_2d * np.invert(binned_up_mask)
        ).all()

        assert (analysis.masked_dataset.mask == binned_up_mask).all()

    def test__phase_can_receive_hyper_image_and_noise_maps(self):
        phase_imaging_7x7 = al.PhaseImaging(
            galaxies=dict(
                lens=al.GalaxyModel(redshift=al.Redshift),
                lens1=al.GalaxyModel(redshift=al.Redshift),
            ),
            hyper_image_sky=al.hyper_data.HyperImageSky,
            hyper_background_noise=al.hyper_data.HyperBackgroundNoise,
            non_linear_class=af.MultiNest,
            phase_name="test_phase",
        )

        instance = phase_imaging_7x7.model.instance_from_vector([0.1, 0.2, 0.3, 0.4])

        assert instance.galaxies[0].redshift == 0.1
        assert instance.galaxies[1].redshift == 0.2
        assert instance.hyper_image_sky.sky_scale == 0.3
        assert instance.hyper_background_noise.noise_scale == 0.4


class TestMakeAnalysis:
    def test__masks_image_and_noise_map_correctly(
        self, phase_imaging_7x7, imaging_7x7, mask_7x7
    ):
        analysis = phase_imaging_7x7.make_analysis(dataset=imaging_7x7, mask=mask_7x7)

        assert (
            analysis.masked_imaging.image.in_2d
            == imaging_7x7.image.in_2d * np.invert(mask_7x7)
        ).all()
        assert (
            analysis.masked_imaging.noise_map.in_2d
            == imaging_7x7.noise_map.in_2d * np.invert(mask_7x7)
        ).all()

    def test___phase_info_is_made(self, phase_imaging_7x7, imaging_7x7, mask_7x7):
        phase_imaging_7x7.make_analysis(dataset=imaging_7x7, mask=mask_7x7)

        file_phase_info = "{}/{}".format(
            phase_imaging_7x7.optimizer.paths.phase_output_path, "phase.info"
        )

        phase_info = open(file_phase_info, "r")

        optimizer = phase_info.readline()
        sub_size = phase_info.readline()
        psf_shape_2d = phase_info.readline()
        positions_threshold = phase_info.readline()
        cosmology = phase_info.readline()

        phase_info.close()

        assert optimizer == "Optimizer = MockNLO \n"
        assert sub_size == "Sub-grid size = 2 \n"
        assert psf_shape_2d == "PSF shape = None \n"
        assert positions_threshold == "Positions Threshold = None \n"
        assert (
            cosmology
            == 'Cosmology = FlatLambdaCDM(name="Planck15", H0=67.7 km / (Mpc s), Om0=0.307, Tcmb0=2.725 K, '
            "Neff=3.05, m_nu=[0.   0.   0.06] eV, Ob0=0.0486) \n"
        )


class TestFit:
    def test__fit_using_imaging(self, imaging_7x7, mask_7x7):
        clean_images()

        phase_imaging_7x7 = al.PhaseImaging(
            non_linear_class=mock_pipeline.MockNLO,
            galaxies=dict(
                lens=al.GalaxyModel(redshift=0.5, light=al.lp.EllipticalSersic),
                source=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
            ),
            phase_name="test_phase_test_fit",
        )

        result = phase_imaging_7x7.run(dataset=imaging_7x7, mask=mask_7x7)
        assert isinstance(result.instance.galaxies[0], al.Galaxy)
        assert isinstance(result.instance.galaxies[0], al.Galaxy)

    def test__figure_of_merit__matches_correct_fit_given_galaxy_profiles(
        self, imaging_7x7, mask_7x7
    ):
        lens_galaxy = al.Galaxy(
            redshift=0.5, light=al.lp.EllipticalSersic(intensity=0.1)
        )

        phase_imaging_7x7 = al.PhaseImaging(
            galaxies=[lens_galaxy],
            cosmology=cosmo.FLRW,
            sub_size=1,
            phase_name="test_phase",
        )

        analysis = phase_imaging_7x7.make_analysis(dataset=imaging_7x7, mask=mask_7x7)
        instance = phase_imaging_7x7.model.instance_from_unit_vector([])
        fit_figure_of_merit = analysis.fit(instance=instance)

        masked_imaging = al.MaskedImaging(imaging=imaging_7x7, mask=mask_7x7)
        tracer = analysis.tracer_for_instance(instance=instance)

        fit = al.FitImaging(masked_imaging=masked_imaging, tracer=tracer)

        assert fit.likelihood == fit_figure_of_merit

    def test__figure_of_merit__includes_hyper_image_and_noise__matches_fit(
        self, imaging_7x7, mask_7x7
    ):
        hyper_image_sky = al.hyper_data.HyperImageSky(sky_scale=1.0)
        hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

        lens_galaxy = al.Galaxy(
            redshift=0.5, light=al.lp.EllipticalSersic(intensity=0.1)
        )

        phase_imaging_7x7 = al.PhaseImaging(
            galaxies=[lens_galaxy],
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
            cosmology=cosmo.FLRW,
            sub_size=4,
            phase_name="test_phase",
        )

        analysis = phase_imaging_7x7.make_analysis(dataset=imaging_7x7, mask=mask_7x7)
        instance = phase_imaging_7x7.model.instance_from_unit_vector([])
        fit_figure_of_merit = analysis.fit(instance=instance)

        mask = phase_imaging_7x7.meta_dataset.mask_with_phase_sub_size_from_mask(
            mask=mask_7x7
        )
        assert mask.sub_size == 4

        masked_imaging = al.MaskedImaging(imaging=imaging_7x7, mask=mask)
        tracer = analysis.tracer_for_instance(instance=instance)
        fit = FitImaging(
            masked_imaging=masked_imaging,
            tracer=tracer,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
        )

        assert fit.likelihood == fit_figure_of_merit

    def test__uses_hyper_fit_correctly(self, instance, result, masked_imaging_7x7):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", result)
        results_collection[0].use_as_hyper_dataset = True
        analysis = al.PhaseImaging.Analysis(
            masked_imaging=masked_imaging_7x7,
            cosmology=cosmo.Planck15,
            results=results_collection,
            image_path="",
        )

        hyper_galaxy = al.HyperGalaxy(
            contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
        )

        instance.galaxies.lens.hyper_galaxy = hyper_galaxy

        fit_likelihood = analysis.fit(instance=instance)

        lens_hyper_image = result.image_galaxy_dict[("galaxies", "lens")]
        source_hyper_image = result.image_galaxy_dict[("galaxies", "source")]

        hyper_model_image = lens_hyper_image + source_hyper_image

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=instance.galaxies.lens.light,
            mass_profile=instance.galaxies.lens.mass,
            hyper_galaxy=hyper_galaxy,
            hyper_model_image=hyper_model_image,
            hyper_galaxy_image=lens_hyper_image,
            hyper_minimum_value=0.0,
        )
        g1 = al.Galaxy(redshift=1.0, light_profile=instance.galaxies.source.light)

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

        fit = FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

        assert (fit_likelihood == fit.likelihood).all()
